#!/usr/bin/perl

# $Id: process-quarantine.pl 897 2005-12-23 03:28:31Z dmorton $

########################################################################
# MAIA MAILGUARD LICENSE v.1.0
#
# Copyright 2004 by Robert LeBlanc <rjl@renaissoft.com>
#                   David Morton   <mortonda@dgrmm.net>
# All rights reserved.
#
# PREAMBLE
#
# This License is designed for users of Maia Mailguard
# ("the Software") who wish to support the Maia Mailguard project by
# leaving "Maia Mailguard" branding information in the HTML output
# of the pages generated by the Software, and providing links back
# to the Maia Mailguard home page.  Users who wish to remove this
# branding information should contact the copyright owner to obtain
# a Rebranding License.
#
# DEFINITION OF TERMS
#
# The "Software" refers to Maia Mailguard, including all of the
# associated PHP, Perl, and SQL scripts, documentation files, graphic
# icons and logo images.
#
# GRANT OF LICENSE
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. The end-user documentation included with the redistribution, if
#    any, must include the following acknowledgment:
#
#    "This product includes software developed by Robert LeBlanc
#    <rjl@renaissoft.com>."
#
#    Alternately, this acknowledgment may appear in the software itself,
#    if and wherever such third-party acknowledgments normally appear.
#
# 4. At least one of the following branding conventions must be used:
#
#    a. The Maia Mailguard logo appears in the page-top banner of
#       all HTML output pages in an unmodified form, and links
#       directly to the Maia Mailguard home page; or
#
#    b. The "Powered by Maia Mailguard" graphic appears in the HTML
#       output of all gateway pages that lead to this software,
#       linking directly to the Maia Mailguard home page; or
#
#    c. A separate Rebranding License is obtained from the copyright
#       owner, exempting the Licensee from 4(a) and 4(b), subject to
#       the additional conditions laid out in that license document.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
########################################################################

use strict;
use Getopt::Long;

# CONFIGURE THIS: Location of your process-quarantine-sub.pl script
my $subroutine = "/var/amavisd/maia/scripts/process-quarantine-sub.pl";

# CONFIGURE THIS: Location to write the lock/PID file
my $pid_file = "/var/amavisd/.process-quarantine.pid";

########################################################################
# End of user-configurable portion.  There should be no need to modify #
# anything below this point.                                           #
########################################################################

my $limit = 0;
my $learn = 0;
my $report = 0;
my $spamonly = 0;
my $hamonly = 0;
my $help = 0;
my $debug = 0;
my $quiet = 0;

GetOptions("limit=i" => \$limit,	# --limit=int or --limit int
           "learn" => \$learn,		# --learn
           "report" => \$report,	# --report
           "spam-only" => \$spamonly,	# --spam-only
           "ham-only" => \$hamonly,	# --ham-only
           "help" => \$help,		# --help
           "debug" => \$debug,		# --debug
           "quiet" => \$quiet);		# --quiet

# Display usage information
if ($help) {
    print "process-quarantine.pl\n" .
          "   --ham-only  : only process confirmed ham (non-spam) items\n" .
          "   --spam-only : only process confirmed spam items\n" .
          "   --learn     : train the Bayes database\n" .
          "   --report    : report spam to Razor2/DCC/Pyzor\n" .
          "   --limit=n   : process items in batches of n at a time\n" .
          "   --help      : display this help text\n" .
          "   --debug     : display detailed debugging information\n" .
          "   --quiet     : display only error messages\n\n";
    exit;
}

# Resolve any debug/quiet conflicts
if ($debug && $quiet) {
    $debug = 0;
    $quiet = 0;
    print "Maia: [process-quarantine] Warning: --debug and --quiet negate each other.\n";
}

# Check for existing lock file, remove any stale locks
if (open(PID_FILE, "< $pid_file\0")) {
    my $pid;
    while (<PID_FILE>) { chomp; $pid = $_ if /^\d+\z/ }
    close(PID_FILE) or die "Maia: [process-quarantine] Can't close file $pid_file: $!";
    if (!defined($pid)) {
        if ($debug) {
            print "Maia: [process-quarantine] Removing invalid lock file...\n";
        }
        unlink($pid_file);
    } elsif (kill 0 => $pid) {
        if (!$quiet) {
            print "Maia: [process-quarantine] Another instance [$pid] is currently running.\n";
        }
        exit;
    } else {
        if ($debug) {
            print "Maia: [process-quarantine] Removing stale lock file [$pid]...\n";
        }
        unlink($pid_file);
    }
}

# Write lock file, store PID
open(PID_FILE, "> $pid_file\0") or die ("Maia: [process-quarantine] Can't write lock file $pid_file");
print PID_FILE "$$";
close(PID_FILE);

# Assemble command-line to call the subroutine
my @cl_array = ($subroutine);
push(@cl_array, "--limit", $limit) if ($limit); # limit
push(@cl_array, "--learn") if ($learn); # train the Bayes database
push(@cl_array, "--report") if ($report); # report spam
push(@cl_array, "--spam-only") if ($spamonly); # process spam only
push(@cl_array, "--ham-only") if ($hamonly); # process ham only
push(@cl_array, "--debug") if ($debug); # turn on debugging mode
push(@cl_array, "--quiet") if ($quiet); # turn on quiet mode

# Call the subroutine until there are no items left to process
my $total_items_processed = 0;
my $total_processing_runs = 0;
my $items_processed = -1;
while ($items_processed != 0) {
    $items_processed = system(@cl_array) / 256;
    $total_items_processed += $items_processed;
    $total_processing_runs++;
    if ($debug) {
        print "Maia: [process-quarantine] sub processed " . $items_processed . " items.\n";
    }
}
if (!$quiet) {
    print "Processing completed: " . $total_items_processed . " items processed in " .
          $total_processing_runs . " runs.\n"
}

unlink($pid_file);

exit;
